/* OggDropXPd - Filename: AUDIO.C
 *
 * Function: Essentially provides all the wave input and output routines.
 *
 *   Copyright (c) 2002 - 2005 John Edwards <john.edwards33@ntlworld.com>
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Library General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Portions, (c) Michael Smith <msmith@labyrinth.net.au>
 * Portions from Vorbize, (c) Kenneth Arnold <kcarnold@yahoo.com>
 * and libvorbis examples, (c) Monty <monty@xiph.org>
 *
 * AIFF/AIFC support from OggSquish, (c) 1994-1996 Monty <xiphmont@xiph.org>
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <io.h>
#include <fcntl.h>
#include <malloc.h>
#include <math.h>

#include "audio.h"
#include "i18n.h"
#include "samplerate.h"
#include "misc.h"

#ifdef HAVE_LIBFLAC
#include "flac.h"
#endif

/* Macros to read header data */
#define READ_U32_LE(buf) \
	(((buf)[3]<<24)|((buf)[2]<<16)|((buf)[1]<<8)|((buf)[0]&0xff))

#define READ_U16_LE(buf) \
	(((buf)[1]<<8)|((buf)[0]&0xff))

#define READ_U32_BE(buf) \
	(((buf)[0]<<24)|((buf)[1]<<16)|((buf)[2]<<8)|((buf)[3]&0xff))

#define READ_U16_BE(buf) \
	(((buf)[0]<<8)|((buf)[1]&0xff))

static unsigned char pcm_guid[16] =
{
	/* (00000001-0000-0010-8000-00aa00389b71) */

	0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 0x00, 0x80, 0x00, 0x00, 0xaa, 0x00, 0x38, 0x9b, 0x71
};

static unsigned char ieee_float_guid[16] =
{
	/* (00000003-0000-0010-8000-00aa00389b71) */

	0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 0x00, 0x80, 0x00, 0x00, 0xaa, 0x00, 0x38, 0x9b, 0x71
};

/* Define the supported formats here */
input_format formats[] = {
	{wav_id, 12, wav_open, wav_close, "wav", N_("WAV file reader")},
	{aiff_id, 12, aiff_open, wav_close, "aiff", N_("AIFF/AIFC file reader")},
#ifdef HAVE_LIBFLAC
	{flac_id,     4, flac_open, flac_close, "flac", N_("FLAC file reader")},
	{oggflac_id, 32, flac_open, flac_close, "ogg", N_("Ogg FLAC file reader")},
#endif
	{NULL, 0, NULL, NULL, NULL, NULL}
};

input_format *open_audio_file(FILE *in, oe_enc_opt *opt)
{
	int j=0;
	unsigned char *buf=NULL;
	int buf_size=0, buf_filled=0;
	int size,ret;

	while (formats[j].id_func)
	{
		size = formats[j].id_data_len;
		if (size >= buf_size)
		{
			buf = realloc(buf, size);
			buf_size = size;
		}

		if (size > buf_filled)
		{
			ret = fread(buf+buf_filled, 1, buf_size-buf_filled, in);
			buf_filled += ret;

			if (buf_filled < size) {
				/* File truncated */
				j++;
				continue;
			}
		}

		if (formats[j].id_func(buf, buf_filled))
		{
			/* ok, we now have something that can handle the file */
			if (formats[j].open_func(in, opt, buf, buf_filled)) {
				free(buf);
				return &formats[j];
			}
		}
		j++;
	}

	free(buf);

	return NULL;
}

static int seek_forward(FILE *in, unsigned int length)
{
       struct stat stb;
       unsigned char buf[1024];
	   unsigned int seek_needed = length;
	   int seeked;
 
       if(fstat(fileno(in), &stb) == 0 && (stb.st_mode & S_IFMT) == S_IFREG)
       {
               if(fseek(in, length, SEEK_CUR) == 0)
                       return 1;
       }
       while(seek_needed > 0) {
			seeked = fread(buf, 1, seek_needed>1024?1024:seek_needed, in);
			if(!seeked)
				return 0; /* Couldn't read more, can't read file */
			else
				seek_needed -= seeked;
       }
       return 1;
}

static int find_wav_chunk(FILE *in, char *type, unsigned int *len, int to_log_file)
{
	unsigned char buf[8];

	while (1)
	{
		if (fread(buf,1,8,in) < 8) /* Suck down a chunk specifier */
		{
			error_handler("Warning: Unexpected EOF in reading WAV header\n");
			return 0; /* EOF before reaching the appropriate chunk */
		}

		if (memcmp(buf, type, 4))
		{
			*len = READ_U32_LE(buf+4);
			if (!seek_forward(in, *len))
				return 0;

			buf[4] = 0;
			if (to_log_file)
				error_handler("Skipping chunk of type \"%s\", length %d\n", buf, *len);
		}
		else
		{
			*len = READ_U32_LE(buf+4);
			return 1;
		}
	}
}

static int find_aiff_chunk(FILE *in, char *type, unsigned int *len)
{
	unsigned char buf[8];
	int restarted = 0;

	while(1)
	{
		if(fread(buf,1,8,in) <8)
		{
			if(!restarted) {
				/* Handle out of order chunks by seeking back to the start
				 * to retry */
				restarted = 1;
				fseek(in, 12, SEEK_SET);
				continue;
			}
			error_handler("Warning: Unexpected EOF in AIFF chunk\n");
			return 0;
		}

		*len = READ_U32_BE(buf+4);

		if(memcmp(buf,type,4))
		{
			if((*len) & 0x1)
				(*len)++;

			if(!seek_forward(in, *len))
				return 0;
		}
		else
			return 1;
	}
}



double read_IEEE80(unsigned char *buf)
{
	int s=buf[0]&0xff;
	int e=((buf[0]&0x7f)<<8)|(buf[1]&0xff);
	double f=((unsigned long)(buf[2]&0xff)<<24)|
		((buf[3]&0xff)<<16)|
		((buf[4]&0xff)<<8) |
		 (buf[5]&0xff);

	if (e==32767)
	{
		if (buf[2]&0x80)
			return HUGE_VAL; /* Really NaN, but this won't happen in reality */
		else
		{
			if (s)
				return -HUGE_VAL;
			else
				return HUGE_VAL;
		}
	}

	f=ldexp(f,32);
	f+= ((buf[6]&0xff)<<24)|
		((buf[7]&0xff)<<16)|
		((buf[8]&0xff)<<8) |
		 (buf[9]&0xff);

	return ldexp(f, e-16446);
}

/* AIFF/AIFC support adapted from the old OggSQUISH application */
int aiff_id(unsigned char *buf, int len)
{
	if (len<12) return 0; /* Truncated file, probably */

	if (memcmp(buf, "FORM", 4))
		return 0;

	if (memcmp(buf+8, "AIF",3))
		return 0;

	if (buf[11]!='C' && buf[11]!='F')
		return 0;

	return 1;
}

static int aiff_permute_matrix[6][6] = 
{
  {0},              /* 1.0 mono   */
  {0,1},            /* 2.0 stereo */
  {0,2,1},          /* 3.0 channel ('wide') stereo */
  {0,1,2,3},        /* 4.0 discrete quadraphonic (WARN) */
  {0,2,1,3,4},      /* 5.0 surround (WARN) */
  {0,1,2,3,4,5},    /* 5.1 surround (WARN)*/
};


int aiff_open(FILE *in, oe_enc_opt *opt, unsigned char *buf, int buflen)
{
	int aifc; /* AIFC or AIFF? */
	unsigned int len;
	unsigned char *buffer;
	unsigned char buf2[8];
	aiff_fmt format;
	aifffile *aiff = malloc(sizeof(aifffile));
	int i;

	if (buf[11]=='C')
		aifc=1;
	else
		aifc=0;

	if (!find_aiff_chunk(in, "COMM", &len))
	{
		error_handler("Warning: No common chunk found in AIFF file\n");
		return 0; /* EOF before COMM chunk */
	}

	if (len < 18)
	{
		error_handler("Warning: Truncated common chunk in AIFF header\n");
		return 0; /* Weird common chunk */
	}

	buffer = alloca(len);

	if (fread(buffer,1,len,in) < len)
	{
		error_handler("Warning: Unexpected EOF in reading AIFF header\n");
		return 0;
	}

	format.channels = READ_U16_BE(buffer);
	format.totalframes = READ_U32_BE(buffer+2);
	format.samplesize = READ_U16_BE(buffer+6);
	format.rate = (int)read_IEEE80(buffer+8);

	aiff->bigendian = 1;

	if (aifc)
	{
		if (len < 22)
		{
			error_handler("Warning: AIFF-C header truncated.\n");
			return 0;
		}

		if (!memcmp(buffer+18, "NONE", 4)) 
		{
			aiff->bigendian = 1;
		}
		else if (!memcmp(buffer+18, "sowt", 4)) 
		{
			aiff->bigendian = 0;
		}
		else
		{
			error_handler("Warning: Can't handle compressed AIFF-C (%c%c%c%c)\n", *(buffer+18), *(buffer+19), *(buffer+20), *(buffer+21));
			return 0; /* Compressed. Can't handle */
		}
	}

	if (!find_aiff_chunk(in, "SSND", &len))
	{
		error_handler("Warning: No SSND chunk found in AIFF file\n");
		return 0; /* No SSND chunk -> no actual audio */
	}

	if (len < 8)
	{
		error_handler("Warning: Corrupted SSND chunk in AIFF header\n");
		return 0; 
	}

	if (fread(buf2,1,8, in) < 8)
	{
		error_handler("Warning: Unexpected EOF reading AIFF header\n");
		return 0;
	}

	format.offset = READ_U32_BE(buf2);
	format.blocksize = READ_U32_BE(buf2+4);

	if ( format.blocksize == 0 && 
		 (format.samplesize == 16 || format.samplesize == 8))
	{
		/* From here on, this is very similar to the wav code. Oh well. */
/*        if(opt->ignorelength)
        {
            format.totalframes = -1;
        }
*/		
		opt->rate = format.rate;
		opt->channels = format.channels;
		opt->read_samples = wav_read; /* Similar enough, so we use the same */
		opt->total_samples_per_channel = format.totalframes;

		aiff->f = in;
		aiff->samplesread = 0;
		aiff->channels = format.channels;
		aiff->samplesize = format.samplesize;
		aiff->totalsamples = format.totalframes;

		if(aiff->channels>3)
			fprintf(stderr,"WARNING: AIFF[-C] files with greater than three channels use\n"
				"speaker locations incompatable with Vorbis suppound definitions.\n"
				"Not performaing channel location mapping.\n");

		opt->readdata = (void *)aiff;

		aiff->channel_permute = malloc(aiff->channels * sizeof(int));
		if (aiff->channels <= 6)
			/* Where we know the mappings, use them. */
			memcpy(aiff->channel_permute, aiff_permute_matrix[aiff->channels-1], 
				sizeof(int) * aiff->channels);
		else
			/* Use a default 1-1 mapping */
			for (i=0; i < aiff->channels; i++)
				aiff->channel_permute[i] = i;

		seek_forward(in, format.offset); /* Swallow some data */
		return 1;
	}
	else
	{
		error_handler("Warning: OggEnc does not support this type of AIFF/AIFC file\n"
				" Must be 8 or 16 bit PCM.\n");
		return 0;
	}
}



int wav_id(unsigned char *buf, int len)
{
	unsigned int flen;
	
	if (len<12) return 0; /* Something screwed up */

	if (memcmp(buf, "RIFF", 4))
		return 0; /* Not wave */

	flen = READ_U32_LE(buf+4); /* We don't use this */

	if (memcmp(buf+8, "WAVE",4))
		return 0; /* RIFF, but not wave */

	return 1;
}

static int wav_permute_matrix[8][8] = 
{
  {0},              /* 1.0 mono   */
  {0,1},            /* 2.0 stereo */
  {0,2,1},          /* 3.0 channel ('wide') stereo */
  {0,1,2,3},        /* 4.0 discrete quadraphonic */
  {0,2,1,3,4},      /* 5.0 surround */
  {0,2,1,4,5,3},    /* 5.1 surround */
  {0,2,1,5,6,4,3},  /* 6.1 surround */
  {0,2,1,6,7,4,5,3} /* 7.1 surround (classic theater 8-track) */
};


int wav_open(FILE *in, oe_enc_opt *opt, unsigned char *oldbuf, int buflen)
{
	unsigned char buf[40];
	unsigned int len;
	int samplesize;
	wav_fmt format;
	wavfile *wav = malloc(sizeof(wavfile));
	int i;

	/* Ok. At this point, we know we have a WAV file. Now we have to detect
	 * whether we support the subtype, and we have to find the actual data
	 * We don't (for the wav reader) need to use the buffer we used to id this
	 * as a wav file (oldbuf)
	 */

	if (!find_wav_chunk(in, "fmt ", &len, opt->logerr))
		return 0; /* EOF */

	if (len < 16) 
	{
		error_handler("Warning: Unrecognised format chunk in WAV header\n");
		return 0; /* Weird format chunk */
	}

	/* A common error is to have a format chunk that is not 16, 18 or 40 bytes
	 * in size.  This is incorrect, but not fatal, so we only warn about 
	 * it instead of refusing to work with the file.  Please, if you
	 * have a program that's creating format chunks of sizes other than
	 * 16, 18 or 40 bytes in size, report a bug to the author.
	 * (40 bytes accommodates WAVEFORMATEXTENSIBLE conforming files.)
	 */
	if (len!=16 && len!=18 && len!=40)
		error_handler("Warning: INVALID format chunk in wav header.\n"
				" Trying to read anyway (may not work)...\n");


//    if(len>40)len=40;

	if (fread(buf,1,len,in) < len) 
	{
		error_handler("Warning: Unexpected EOF in reading WAV header\n");
		return 0;
	}

	format.format =      READ_U16_LE(buf); 
	format.channels =    READ_U16_LE(buf+2); 
	format.samplerate =  READ_U32_LE(buf+4);
	format.bytespersec = READ_U32_LE(buf+8);
	format.align =       READ_U16_LE(buf+12);
	format.samplesize =  READ_U16_LE(buf+14);

    if(format.format == -2) /* WAVE_FORMAT_EXTENSIBLE */
    {
      if(len<40)
      {
        error_handler("ERROR: Extended WAV format header invalid (too small)\n");
        return 0;
      }

      format.mask = READ_U32_LE(buf+20);
      /* warn the user if the format mask is not a supported/expected type */
      switch(format.mask){
      case 1539: /* 4.0 using side surround instead of back */
        error_handler("WARNING: WAV file uses side surround instead of rear for quadraphonic;\n"
                "remapping side speakers to rear in encoding.\n");
        break;
      case 1551: /* 5.1 using side instead of rear */
        error_handler("WARNING: WAV file uses side surround instead of rear for 5.1;\n"
                "remapping side speakers to rear in encoding.\n");
        break;
      case 319:  /* 6.1 using rear instead of side */
        error_handler("WARNING: WAV file uses rear surround instead of side for 6.1;\n"
                "remapping rear speakers to side in encoding.\n");
        break;
      case 255:  /* 7.1 'Widescreen' */
        error_handler("WARNING: WAV file is a 7.1 'Widescreen' channel mapping;\n"
                "remapping speakers to Vorbis 7.1 format.\n");
        break;
      case 0:    /* default/undeclared */
      case 1:    /* mono */
      case 3:    /* stereo */
      case 51:   /* quad */
      case 55:   /* 5.0 */
      case 63:   /* 5.1 */
      case 1807: /* 6.1 */
      case 1599: /* 7.1 */
        break;
      default:
        error_handler("WARNING: Unknown WAV surround channel mask: %d\n"
                "blindly mapping speakers using default SMPTE/ITU ordering.\n",
                format.mask);
        break;
      }
      format.format = READ_U16_LE(buf+24);
    }

    if(!find_wav_chunk(in, "data", &len, opt->logerr))
        return 0; /* EOF */

    if(format.format == 1)
    {
        samplesize = format.samplesize/8;
        opt->read_samples = wav_read;
    }
    else if(format.format == 3)
    {
        samplesize = 4;
        opt->read_samples = wav_ieee_read;
    }
	else 
	{
		error_handler("ERROR: Wav file is unsupported type (must be standard PCM\n"
			" or type 3 floating point PCM\n");
		return 0;
	}

	if(format.align != format.channels * samplesize) {
		/* This is incorrect according to the spec. Warn loudly, then ignore
		 * this value.
		 */
		error_handler("Warning: WAV 'block alignment' value is incorrect, ignoring.\n" 
		    "The software that created this file is incorrect.\n");
	}


	if(format.samplesize == samplesize * 8 && 
		(format.samplesize == 32 || format.samplesize == 24 ||
		 format.samplesize == 16 || format.samplesize == 8)) 
	{
		/* OK, good - we have a supported format,
		 * now we want to find the size of the file
		 */
		opt->rate = format.samplerate;
		opt->channels = format.channels;

		wav->f = in;
		wav->samplesread = 0;
		wav->bigendian = 0;
		wav->channels = format.channels; /* This is in several places. The price
											of trying to abstract stuff. */
		wav->samplesize = format.samplesize;

/*        if(opt->ignorelength)
        {
            opt->total_samples_per_channel = -1;
        }
        else
        {
*/          if(len)
            {
                opt->total_samples_per_channel =
                    len/(format.channels*samplesize);
            }
            else
            {
                long pos;
                pos = ftell(in);
                if(fseek(in, 0, SEEK_END) == -1)
                {
                    opt->total_samples_per_channel = 0; /* Give up */
                }
                else
                {
                    opt->total_samples_per_channel = (ftell(in) - pos)/
                        (format.channels*samplesize);
                    fseek(in,pos, SEEK_SET);
                }
            }
/*        }
*/
		wav->totalsamples = opt->total_samples_per_channel;

		opt->readdata = (void *)wav;

		wav->channel_permute = malloc(wav->channels * sizeof(int));
		if (wav->channels <= 8)
			/* Where we know the mappings, use them. */
			memcpy(wav->channel_permute, wav_permute_matrix[wav->channels-1],
					sizeof(int) * wav->channels);
		else
			/* Use a default 1-1 mapping */
			for (i=0; i < wav->channels; i++)
				wav->channel_permute[i] = i;

		return 1;
	}
	else
	{
		error_handler("ERROR: Wav file is unsupported subformat (must be 8, 16, 24 or 32 bit PCM\n"
				"or floating point PCM)\n");
		return 0;
	}
}


long wav_read(void *in, float **buffer, int samples)
{
	wavfile *f = (wavfile *)in;
	int sampbyte = f->samplesize / 8;
	signed char *buf = alloca(samples*sampbyte*f->channels);
	long bytes_read = fread(buf, 1, samples*sampbyte*f->channels, f->f);
	int i,j;
	long realsamples;
	int *ch_permute = f->channel_permute;

	if (f->totalsamples && f->samplesread +
			 bytes_read/(sampbyte*f->channels) > f->totalsamples) {
		bytes_read = sampbyte*f->channels*(f->totalsamples - f->samplesread);
	}

	realsamples = bytes_read/(sampbyte*f->channels);
	f->samplesread += realsamples;

	if (f->samplesize==8)
	{
		unsigned char *bufu = (unsigned char *)buf;
		for (i = 0; i < realsamples; i++)
		{
			for (j=0; j < f->channels; j++)
			{
				buffer[j][i]=((int)(bufu[i*f->channels + ch_permute[j]])-128)/128.0f;
			}
		}
	}
	else if (f->samplesize==16)
	{
		if(!f->bigendian)
		{
			for(i = 0; i < realsamples; i++)
			{
				for(j=0; j < f->channels; j++)
				{
					buffer[j][i] = ((buf[i*2*f->channels + 2*ch_permute[j] + 1]<<8) |
							(buf[i*2*f->channels + 2*ch_permute[j]] & 0xff))/32768.0f;
				}
			}
		}
		else
		{
			for(i = 0; i < realsamples; i++)
			{
				for(j=0; j < f->channels; j++)
				{
					buffer[j][i] = ((buf[i*2*f->channels + 2*ch_permute[j]]<<8) |
							(buf[i*2*f->channels + 2*ch_permute[j] + 1] & 0xff))/32768.0f;
				}
			}
		}
	}
	else if (f->samplesize==24)
	{
		if (!f->bigendian)
		{
			for (i = 0; i < realsamples; i++)
			{
				for (j=0; j < f->channels; j++)
				{
					buffer[j][i] = 	 	  ((buf[i*3*f->channels + 3*ch_permute[j] + 2] << 16) |
						(((unsigned char *)buf)[i*3*f->channels + 3*ch_permute[j] + 1] << 8) |
						(((unsigned char *)buf)[i*3*f->channels + 3*ch_permute[j]] & 0xff)) 
						/ 8388608.0f;
				}
			}
		}
		else
		{
			error_handler("Big endian 24 bit PCM data is not currently "
				"supported, aborting.\n");
			return 0;
		}
	}
	else if (f->samplesize==32)
	{
		if (!f->bigendian)
		{
			for (i = 0; i < realsamples; i++)
			{
				for (j=0; j < f->channels; j++)
					buffer[j][i] = 		  ((buf[i*4*f->channels + 4*ch_permute[j] + 3] << 24) |
						(((unsigned char *)buf)[i*4*f->channels + 4*ch_permute[j] + 2] << 16) |
						(((unsigned char *)buf)[i*4*f->channels + 4*ch_permute[j] + 1] << 8) |
						(((unsigned char *)buf)[i*4*f->channels + 4*ch_permute[j]] & 0xff)) 
						/ 2147483648.0f;
			}
		}
		else {
			error_handler("Big endian 32 bit PCM data is not currently "
				"supported, aborting.\n");
			return 0;
		}
	}
	else {
		error_handler("Internal error: attempt to read unsupported "
			"bitdepth %d\n", f->samplesize);
		return 0;
	}

	return realsamples;
}

long wav_ieee_read(void *in, float **buffer, int samples)
{
	wavfile *f = (wavfile *)in;
	float *buf = alloca(samples*4*f->channels); /* de-interleave buffer */
	long bytes_read = fread(buf,1,samples*4*f->channels, f->f);
	int i,j;
	long realsamples;


	if (f->totalsamples && f->samplesread + 
			 bytes_read/(4*f->channels) > f->totalsamples)
		bytes_read = 4*f->channels*(f->totalsamples - f->samplesread);
	realsamples = bytes_read/(4*f->channels);
	f->samplesread += realsamples;

	for (i=0; i < realsamples; i++)
		for (j=0; j < f->channels; j++)
			buffer[j][i] = buf[i*f->channels + f->channel_permute[j]];

	return realsamples;
}


void wav_close(void *info)
{
	wavfile *f = (wavfile *)info;
	free(f->channel_permute);

	free(f);
}

int raw_open(FILE *in, oe_enc_opt *opt)
{
	wav_fmt format; /* fake wave header ;) */
	wavfile *wav = malloc(sizeof(wavfile));
	int i;

	/* construct fake wav header ;) */
	format.format =      2; 
	format.channels =    opt->channels;
	format.samplerate =  opt->rate;
	format.samplesize =  opt->samplesize;
	format.bytespersec = opt->channels * opt->rate * opt->samplesize / 8;
	format.align =       format.bytespersec;
	wav->f =             in;
	wav->samplesread =   0;
	wav->bigendian =     opt->endianness;
	wav->channels =      format.channels;
	wav->samplesize =    opt->samplesize;
	wav->totalsamples =  0;
	wav->channel_permute = malloc(wav->channels * sizeof(int));
	for (i=0; i < wav->channels; i++)
		wav->channel_permute[i] = i;

/*	if (opt->format == 3)
		opt->read_samples = wav_ieee_read;
	else */
		opt->read_samples = wav_read;
	opt->readdata = (void *)wav;
	opt->total_samples_per_channel = 0; /* raw mode, don't bother */
	return 1;
}

typedef struct {
	audio_read_func real_reader;
	SRC_STATE *src_state;
	SRC_DATA  src_data;
	void *real_readdata;
	float **buffs;
	float *in_sample_buffer;
	float *out_sample_buffer;
	long read_size;
	unsigned int converter;
	int channels;
} resampler;

static long read_resampled(void *d, float **buffer, int samples)
{
	resampler *rs = d;
	int       error;
	long      in_samples;
	int       i, ch;

	rs->src_data.end_of_input = 0;
	in_samples = rs->real_reader(rs->real_readdata, rs->buffs, rs->read_size);

	for (ch = 0; ch < rs->channels; ch++)
		for (i = 0; i < in_samples; i++)
			rs->in_sample_buffer[(i*rs->channels)+ch] = rs->buffs[ch][i];

	rs->src_data.input_frames = in_samples;
	rs->src_data.data_in = rs->in_sample_buffer;

	if (rs->src_data.input_frames < rs->read_size)
		rs->src_data.end_of_input = 1;

	if ((error = src_process (rs->src_state, &rs->src_data))) {
		error_handler("Error : %s\n", src_strerror (error));
		return -1;
	}

	if (rs->src_data.end_of_input && rs->src_data.output_frames_gen == 0)
		return 0;

	for (i=0; i < rs->src_data.output_frames_gen; i++)
		for (ch=0; ch < rs->channels; ch++)
			buffer[ch][i] = rs->out_sample_buffer[i * rs->channels + ch];

	return rs->src_data.output_frames_gen;
}

int setup_resample(oe_enc_opt *opt)
{
	resampler *rs = calloc(1, sizeof(resampler));
	int c;
	int error;
	int output_len;

	rs->src_state = calloc(1, sizeof(rs->src_state));
	rs->real_reader = opt->read_samples;
	rs->buffs = malloc(sizeof(float *) * opt->channels);
	rs->real_readdata = opt->readdata;
	rs->channels = opt->channels;
	rs->converter = opt->converter;
	rs->src_data.src_ratio = opt->ratio;
	rs->src_data.input_frames = 0;
	rs->src_data.end_of_input = 0;

	if (rs->src_data.src_ratio >= 1.0) {
		rs->read_size = ((int)(floor(READSIZE / rs->src_data.src_ratio)) >> 1 << 1);
		output_len = READSIZE;
	}
	else {
		rs->read_size = READSIZE;
		output_len = (int)floor(READSIZE * rs->src_data.src_ratio);
	}

	/* Reduce input_len by 10 so output is longer than necessary. */
	rs->read_size -= 10;
		
	/* Initialize the sample rate converter. */
	if ((rs->src_state = src_new (rs->converter, rs->channels, &error)) == NULL) {
		error_handler("Error : src_new() failed : %s.\n", strerror (error)) ;
		return -1;
	}

	for (c = 0; c < rs->channels; c++)
		rs->buffs[c] = malloc(rs->read_size * sizeof(float));
	rs->in_sample_buffer = malloc(sizeof(float) * rs->read_size * rs->channels);
	rs->out_sample_buffer = malloc(sizeof(float) * output_len * rs->channels);
	rs->src_data.data_out = rs->out_sample_buffer;
	rs->src_data.output_frames = output_len;

	opt->read_samples = read_resampled;
	opt->readdata = rs;
	if (opt->total_samples_per_channel)
		opt->total_samples_per_channel = (int)((float)opt->total_samples_per_channel * 
			((float)opt->resamplefreq/(float)opt->rate));
	opt->rate = opt->resamplefreq;

	return 0;
}

void clear_resample(oe_enc_opt *opt)
{
	resampler *rs = opt->readdata;
	int i;

	opt->read_samples = rs->real_reader;
	opt->readdata = rs->real_readdata;

	for (i = 0; i < rs->channels; i++)
		free(rs->buffs[i]);

	free(rs->buffs);
	free(rs->in_sample_buffer);
	free(rs->out_sample_buffer);
	rs->src_state = src_delete (rs->src_state);
	free(rs->src_state);

	free(rs);
}

typedef struct {
	audio_read_func real_reader;
	void *real_readdata;
	int channels;
	float scale_factor;
} scaler;

static long read_scaler(void *data, float **buffer, int samples)
{
	scaler *d = data;
	long in_samples = d->real_reader(d->real_readdata, buffer, samples);
	int i,j;

	for (i=0; i < d->channels; i++) {
		for (j=0; j < in_samples; j++) {
			buffer[i][j] *= d->scale_factor;
		}
	}

	return in_samples;
}


void setup_scaler(oe_enc_opt *opt, float scale)
{
	scaler *d = calloc(1, sizeof(scaler));

	d->real_reader = opt->read_samples;
	d->real_readdata = opt->readdata;

	opt->read_samples = read_scaler;
	opt->readdata = d;
	d->channels = opt->channels;
	d->scale_factor = scale;
}

void clear_scaler(oe_enc_opt *opt)
{
	scaler *d = opt->readdata;

	opt->read_samples = d->real_reader;
	opt->readdata = d->real_readdata;

	free(d);
}

typedef struct {
	audio_read_func real_reader;
	void *real_readdata;
	float **bufs;
} downmix;

static long read_downmix(void *data, float **buffer, int samples)
{
	downmix *d = data;
	long in_samples = d->real_reader(d->real_readdata, d->bufs, samples);
	int i;

	for (i=0; i < in_samples; i++) {
		buffer[0][i] = (d->bufs[0][i] + d->bufs[1][i])*0.5f;
	}

	return in_samples;
}

void setup_downmix(oe_enc_opt *opt)
{
	downmix *d = calloc(1, sizeof(downmix));

	if (opt->channels != 2) {
		error_handler("Internal error! Please report this bug.\n");
		return;
	}
    
	d->bufs = malloc(2 * sizeof(float *));
	d->bufs[0] = malloc(4096 * sizeof(float));
	d->bufs[1] = malloc(4096 * sizeof(float));
	d->real_reader = opt->read_samples;

	d->real_readdata = opt->readdata;

	opt->read_samples = read_downmix;
	opt->readdata = d;

	opt->channels = 1;
}

void clear_downmix(oe_enc_opt *opt)
{
	downmix *d = opt->readdata;

	opt->read_samples = d->real_reader;
	opt->readdata = d->real_readdata;
	opt->channels = 2; /* other things in cleanup rely on this */

	free(d->bufs[0]);
	free(d->bufs[1]);
	free(d->bufs);
	free(d);
}

/*
 * W A V E   O U T P U T
 */

audio_file *open_output_audio_file(char *infile, int samplerate, int channels,
                            int outputFormat, int fileType)
{
	audio_file *aufile = malloc(sizeof(audio_file));

	aufile->outputFormat = outputFormat;

	aufile->samplerate = samplerate;
	aufile->channels = channels;
	aufile->samples = 0;
	aufile->fileType = fileType;

	switch (outputFormat) {
		case WAV_FMT_8BIT:
			aufile->bits_per_sample = 8;
			break;
		case WAV_FMT_16BIT:
			aufile->bits_per_sample = 16;
			break;
		case WAV_FMT_24BIT:
			aufile->bits_per_sample = 24;
			break;
		case WAV_FMT_32BIT:
		case WAV_FMT_FLOAT:
			aufile->bits_per_sample = 32;
			break;
		default:
			if (aufile)
				free(aufile);
			return NULL;
	}

	aufile->sndfile = fopen(infile, "wb");

	if (aufile->sndfile == NULL) {
		if (aufile)
			free(aufile);
		return NULL;
	}

	if (aufile->fileType == OUTPUT_WAV) {
		if (aufile->channels > 2)
			write_wav_extensible_header(aufile);
		else
			write_wav_header(aufile);
	}
	return aufile;
}

int write_audio_file(audio_file *aufile, void *sample_buffer, int samples)
{
	switch (aufile->outputFormat) {
		case WAV_FMT_8BIT:
			return write_audio_8bit(aufile, sample_buffer, samples);
		case WAV_FMT_16BIT:
			return write_audio_16bit(aufile, sample_buffer, samples);
		case WAV_FMT_24BIT:
			return write_audio_24bit(aufile, sample_buffer, samples);
		case WAV_FMT_32BIT:
			return write_audio_32bit(aufile, sample_buffer, samples);
		case WAV_FMT_FLOAT:
			return write_audio_float(aufile, sample_buffer, samples);
		default:
			return 0;
	}

	return 0;
}

void close_audio_file(audio_file *aufile)
{
	if (aufile->fileType == OUTPUT_WAV) {
		fseek(aufile->sndfile, 0, SEEK_SET);
		if (aufile->channels > 2)
			write_wav_extensible_header(aufile);
		else
			write_wav_header(aufile);
	}

	fclose(aufile->sndfile);

	if (aufile)
		free(aufile);
}

static int write_wav_header(audio_file *aufile)
{
	unsigned char  header[44],
	               *p = header;
	unsigned int   bytes = (aufile->bits_per_sample + 7) / 8;
	unsigned long  data_size = aufile->samples * bytes;
	unsigned long  word32;

	*p++ = 'R';
	*p++ = 'I';
	*p++ = 'F';
	*p++ = 'F';                                            // "RIFF" label

	word32 = data_size + (44 - 8) < MAXWAVESIZE ?
		data_size + (44 - 8)  :  MAXWAVESIZE;
	*p++ = (unsigned char)(word32 >>  0);
	*p++ = (unsigned char)(word32 >>  8);
	*p++ = (unsigned char)(word32 >> 16);
	*p++ = (unsigned char)(word32 >> 24);                  // Size of next chunk

	*p++ = 'W';
	*p++ = 'A';
	*p++ = 'V';
	*p++ = 'E';                                            // "WAVE" label

	*p++ = 'f';
	*p++ = 'm';
	*p++ = 't';
	*p++ = ' ';                                            // "fmt " label

	*p++ = 0x10;
	*p++ = 0x00;
	*p++ = 0x00;
	*p++ = 0x00;                                           // length of the PCM declaration = 2+2+4+4+2+2

	if (aufile->outputFormat == WAV_FMT_FLOAT) {
		*p++ = 0x03;
		*p++ = 0x00;                                   // ACM type 0x003 = uncompressed linear floating point
	}
	else {
		*p++ = 0x01;
		*p++ = 0x00;                                   // ACM type 0x0001 = uncompressed linear PCM
	}

	*p++ = (unsigned char)(aufile->channels >> 0);
	*p++ = (unsigned char)(aufile->channels >> 8);         // Channels

	word32 = (unsigned long) (aufile->samplerate + 0.5);
	*p++ = (unsigned char)(word32 >>  0);
	*p++ = (unsigned char)(word32 >>  8);
	*p++ = (unsigned char)(word32 >> 16);
	*p++ = (unsigned char)(word32 >> 24);                  // Sample frequency

	word32 *= bytes * aufile->channels;
	*p++ = (unsigned char)(word32 >>  0);
	*p++ = (unsigned char)(word32 >>  8);
	*p++ = (unsigned char)(word32 >> 16);
	*p++ = (unsigned char)(word32 >> 24);                  // Bytes per second in the data stream

	word32 = bytes * aufile->channels;
	*p++ = (unsigned char)(word32 >>  0);
	*p++ = (unsigned char)(word32 >>  8);                  // Bytes per sample time

	*p++ = (unsigned char)(aufile->bits_per_sample >> 0);
	*p++ = (unsigned char)(aufile->bits_per_sample >> 8);  // Bits per single sample

	*p++ = 'd';
	*p++ = 'a';
	*p++ = 't';
	*p++ = 'a';                                            // "data" label

	word32 = data_size < MAXWAVESIZE ? data_size : MAXWAVESIZE;
	*p++ = (unsigned char)(word32 >>  0);
	*p++ = (unsigned char)(word32 >>  8);
	*p++ = (unsigned char)(word32 >> 16);
	*p++ = (unsigned char)(word32 >> 24);                  // Real length of PCM data

	return fwrite(header, sizeof(header), 1, aufile->sndfile);
}

static int write_wav_extensible_header(audio_file *aufile)
{
	unsigned char header[68];
	unsigned char* p = header;
	unsigned int bytes = (aufile->bits_per_sample + 7) / 8;
	float data_size = (float)bytes * aufile->samples;
	unsigned long channelMask;
	unsigned long word32;

	if (aufile->channels == 3)
		channelMask = 7;
	else if (aufile->channels == 4)
		channelMask = 51;
	else if (aufile->channels == 5)
		channelMask = 55;
	else if (aufile->channels == 6)
		channelMask = 63;
	else if (aufile->channels == 7)
		channelMask = 1807;
	else if (aufile->channels == 8)
		channelMask = 1599;
	else
		channelMask = 0;

	*p++ = 'R'; *p++ = 'I'; *p++ = 'F'; *p++ = 'F';

	word32 = (data_size + (68 - 8) < (float)MAXWAVESIZE) ?
        	(unsigned long)data_size + (68 - 8)  :  (unsigned long)MAXWAVESIZE;
	*p++ = (unsigned char)(word32 >>  0);
	*p++ = (unsigned char)(word32 >>  8);
	*p++ = (unsigned char)(word32 >> 16);
	*p++ = (unsigned char)(word32 >> 24);

	*p++ = 'W'; *p++ = 'A'; *p++ = 'V'; *p++ = 'E';

	*p++ = 'f'; *p++ = 'm'; *p++ = 't'; *p++ = ' ';

	*p++ = /*0x10*/0x28; *p++ = 0x00; *p++ = 0x00; *p++ = 0x00;

	/* WAVE_FORMAT_EXTENSIBLE */
	*p++ = 0xFE; *p++ = 0xFF;

	*p++ = (unsigned char)(aufile->channels >> 0);
	*p++ = (unsigned char)(aufile->channels >> 8);

	word32 = (unsigned long)(aufile->samplerate + 0.5);
	*p++ = (unsigned char)(word32 >>  0);
	*p++ = (unsigned char)(word32 >>  8);
	*p++ = (unsigned char)(word32 >> 16);
	*p++ = (unsigned char)(word32 >> 24);

	word32 = aufile->samplerate * bytes * aufile->channels;
	*p++ = (unsigned char)(word32 >>  0);
	*p++ = (unsigned char)(word32 >>  8);
	*p++ = (unsigned char)(word32 >> 16);
	*p++ = (unsigned char)(word32 >> 24);

	word32 = bytes * aufile->channels;
	*p++ = (unsigned char)(word32 >>  0);
	*p++ = (unsigned char)(word32 >>  8);

	*p++ = (unsigned char)(aufile->bits_per_sample >> 0);
	*p++ = (unsigned char)(aufile->bits_per_sample >> 8);

	/* cbSize */
	*p++ = (unsigned char)(22);
	*p++ = (unsigned char)(0);

	/* WAVEFORMATEXTENSIBLE */

	/* wValidBitsPerSample */
	*p++ = (unsigned char)(aufile->bits_per_sample >> 0);
	*p++ = (unsigned char)(aufile->bits_per_sample >> 8);

	/* dwChannelMask */
	word32 = channelMask;
	*p++ = (unsigned char)(word32 >>  0);
	*p++ = (unsigned char)(word32 >>  8);
	*p++ = (unsigned char)(word32 >> 16);
	*p++ = (unsigned char)(word32 >> 24);

	/* SubFormat */
	if (aufile->outputFormat == WAV_FMT_FLOAT)
	{
		/* KSDATAFORMAT_SUBTYPE_IEEE_FLOAT: 00000003-0000-0010-8000-00aa00389b71 */
		*p++ = 0x03;
		*p++ = 0x00;
		*p++ = 0x00;
		*p++ = 0x00;
		*p++ = 0x00; *p++ = 0x00; *p++ = 0x10; *p++ = 0x00; *p++ = 0x80; *p++ = 0x00;
		*p++ = 0x00; *p++ = 0xaa; *p++ = 0x00; *p++ = 0x38; *p++ = 0x9b; *p++ = 0x71;
	} else {
		/* KSDATAFORMAT_SUBTYPE_PCM: 00000001-0000-0010-8000-00aa00389b71 */
		*p++ = 0x01;
		*p++ = 0x00;
		*p++ = 0x00;
		*p++ = 0x00;
		*p++ = 0x00; *p++ = 0x00; *p++ = 0x10; *p++ = 0x00; *p++ = 0x80; *p++ = 0x00;
		*p++ = 0x00; *p++ = 0xaa; *p++ = 0x00; *p++ = 0x38; *p++ = 0x9b; *p++ = 0x71;
	}

	/* end WAVEFORMATEXTENSIBLE */

	*p++ = 'd'; *p++ = 'a'; *p++ = 't'; *p++ = 'a';

	word32 = data_size < MAXWAVESIZE ?
        	(unsigned long)data_size : (unsigned long)MAXWAVESIZE;
	*p++ = (unsigned char)(word32 >>  0);
	*p++ = (unsigned char)(word32 >>  8);
	*p++ = (unsigned char)(word32 >> 16);
	*p++ = (unsigned char)(word32 >> 24);

	return fwrite(header, sizeof(header), 1, aufile->sndfile);
}

static int write_audio_8bit(audio_file *aufile, void *sample_buffer, unsigned int samples)
{
	int              ret;
	unsigned int     i;
	unsigned char    *sample_buffer8 = (unsigned char*)sample_buffer;
	unsigned char    *data = malloc(samples*aufile->bits_per_sample*sizeof(char)/8);

	aufile->samples += samples;

	for (i = 0; i < samples; i++)
		data[i] = (sample_buffer8[i]+128) & 0xFF;

	ret = fwrite(data, samples, aufile->bits_per_sample/8, aufile->sndfile);

	if (data)
		free(data);

	return ret;
}

static int write_audio_16bit(audio_file *aufile, void *sample_buffer, unsigned int samples)
{
	int          ret;
	unsigned int i;
	short        *sample_buffer16 = (short*)sample_buffer;
	char         *data = malloc(samples*aufile->bits_per_sample*sizeof(char)/8);

	aufile->samples += samples;

	for (i = 0; i < samples; i++) {
		data[i*2] = (char)(sample_buffer16[i] & 0xFF);
		data[i*2+1] = (char)((sample_buffer16[i] >> 8) & 0xFF);
	}

	ret = fwrite(data, samples, aufile->bits_per_sample/8, aufile->sndfile);

	if (data)
		free(data);

	return ret;
}

static int write_audio_24bit(audio_file *aufile, void *sample_buffer, unsigned int samples)
{
	int          ret;
	unsigned int i;
	long         *sample_buffer24 = (long*)sample_buffer;
	char         *data = malloc(samples*aufile->bits_per_sample*sizeof(char)/8);

	aufile->samples += samples;

	for (i = 0; i < samples; i++) {
		data[i*3] = (char)(sample_buffer24[i] & 0xFF);
		data[i*3+1] = (char)((sample_buffer24[i] >> 8) & 0xFF);
		data[i*3+2] = (char)((sample_buffer24[i] >> 16) & 0xFF);
	}

	ret = fwrite(data, samples, aufile->bits_per_sample/8, aufile->sndfile);

	if (data)
		free(data);

	return ret;
}

static int write_audio_32bit(audio_file *aufile, void *sample_buffer, unsigned int samples)
{
	int          ret;
	unsigned int i;
	long         *sample_buffer32 = (long*)sample_buffer;
	char         *data = malloc(samples*aufile->bits_per_sample*sizeof(char)/8);

	aufile->samples += samples;

	for (i = 0; i < samples; i++) {
		data[i*4] = (char)(sample_buffer32[i] & 0xFF);
		data[i*4+1] = (char)((sample_buffer32[i] >> 8) & 0xFF);
		data[i*4+2] = (char)((sample_buffer32[i] >> 16) & 0xFF);
		data[i*4+3] = (char)((sample_buffer32[i] >> 24) & 0xFF);
	}

	ret = fwrite(data, samples, aufile->bits_per_sample/8, aufile->sndfile);

	if (data)
		free(data);

	return ret;
}

static int write_audio_float(audio_file *aufile, void *sample_buffer, unsigned int samples)
{
	int           ret;
	unsigned int  i;
	float         *sample_buffer_f = (float*)sample_buffer;
	unsigned char *data = malloc(samples*aufile->bits_per_sample*sizeof(char)/8);

	aufile->samples += samples;

	for (i = 0; i < samples; i++) {
		int   exponent,
		      mantissa,
		      negative = 0;
		float in = sample_buffer_f[i];

		data[i*4] = 0;
		data[i*4+1] = 0;
		data[i*4+2] = 0;
		data[i*4+3] = 0;
		if (in == 0.f)
			continue;

		if (in < 0.0) {
			in *= -1.0;
			negative = 1;
		}
		in = (float)frexp(in, &exponent);
		exponent += 126;
		in *= (float)0x1000000;
		mantissa = (((int)in) & 0x7FFFFF);

		if (negative)
			data[i*4+3] |= 0x80;

		if (exponent & 0x01)
			data[i*4+2] |= 0x80;

		data[i*4] = mantissa & 0xFF;
		data[i*4+1] = (mantissa >> 8) & 0xFF;
		data[i*4+2] |= (mantissa >> 16) & 0x7F;
		data[i*4+3] |= (exponent >> 1) & 0x7F;
	}

	ret = fwrite(data, samples, aufile->bits_per_sample/8, aufile->sndfile);

	if (data)
		free(data);

	return ret;
}

void* output_to_PCM(float **input, void *sample_buffer, int channels,
                    int samples, int format)
{
	int permute[8][8] = {{0}, {0,1}, {0,2,1}, {0,1,2,3}, {0,2,1,3,4}, {0,2,1,5,3,4}, {0,2,1,6,5,3,4}, {0,2,1,6,7,5,3,4}};

	int             i,j,k;

	char            *char_sample_buffer = (char*)sample_buffer;
	short           *short_sample_buffer = (short*)sample_buffer;
	int             *int_sample_buffer = (int*)sample_buffer;
	float           *float_sample_buffer = (float*)sample_buffer;

	/*
	 * Copy output to a standard PCM buffer
	 */
	if (channels > 2 && channels < 9) {
		switch (format) {
		case WAV_FMT_8BIT:
			for (j = 0; j < channels; j++) {
				for (i = 0; i < samples; i++) {
					char_sample_buffer[(i*channels)+j] = input[permute[channels-1][j]][i];
				}
			}
			break;
		case WAV_FMT_16BIT:
			for (j = 0; j < channels; j++) {
				for (i = 0; i < samples; i++) {
					short_sample_buffer[(i*channels)+j] = input[permute[channels-1][j]][i];
				}
			}
			break;
		case WAV_FMT_24BIT:
		case WAV_FMT_32BIT:
			for (j = 0; j < channels; j++) {
				for (i = 0; i < samples; i++) {
					int_sample_buffer[(i*channels)+j] = input[permute[channels-1][j]][i];
				}
			}
			break;
		case WAV_FMT_FLOAT:
			for (j = 0; j < channels; j++) {
				for (i = 0; i < samples; i++) {
					float_sample_buffer[(i*channels)+j] = input[permute[channels-1][j]][i];
				}
			}
			break;
		}
	}
	else
	{
		switch (format) {
		case WAV_FMT_8BIT:
			for (j = 0; j < channels; j++) {
				for (i = 0; i < samples; i++) {
					char_sample_buffer[(i*channels)+j] = input[j][i];
				}
			}
			break;
		case WAV_FMT_16BIT:
			for (j = 0; j < channels; j++) {
				for (i = 0; i < samples; i++) {
					short_sample_buffer[(i*channels)+j] = input[j][i];
				}
			}
			break;
		case WAV_FMT_24BIT:
		case WAV_FMT_32BIT:
			for (j = 0; j < channels; j++) {
				for (i = 0; i < samples; i++) {
					int_sample_buffer[(i*channels)+j] = input[j][i];
				}
			}
			break;
		case WAV_FMT_FLOAT:
			for (j = 0; j < channels; j++) {
				for (i = 0; i < samples; i++) {
					float_sample_buffer[(i*channels)+j] = input[j][i];
				}
			}
			break;
		}
	}
	return sample_buffer;
}

/*
 * end of audio.c
 */

