/* OggDropXPd - Filename: DECODE.C
 * 
 * Function: ogg decoder and player using vorbisfile and VorbisGain.
 *
 *   Copyright (c) 2002 - 2005 John Edwards <john.edwards33@ntlworld.com>
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Library General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* Takes a vorbis bitstream from input ogg files and plays them through
 * the soundcard using Windows Audio wave mapping, or decodes to file in
 * 8, 16, 24 or 32 bit integer, or 32 bit float format. LibSndFile is
 * used for writing out the wave files. Multiple input files are permitted,
 * with wildcard use. Use is made of ReplayGain tags during playback and 
 * decoding, if requested and if the correct tags are present. Absence of
 * the tags, even if their use is requested, does not cause an error; 
 * decoding/playback proceeds without gain adjustment.
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#include <windows.h>
#include <malloc.h>

#include <ogg/os_types.h>
#include <vorbis/codec.h>
#include <vorbis/vorbisfile.h>

#include "audio.h"
#include "encode.h"
#include "misc.h"
#include "wave_out.h"           // included for WIN_AUDIO
#include "vcedit.h"             // included for vorbis tag reading routines
#include "dither.h"             // included for dither functions

#define READ_SIZE 4096          // length in samples to be read

#define SQRTHALF  0.70710678118654752440084436210485

#define ROUND64(x)   ( doubletmp = (x) + Dither.Add + (ogg_int64_t)0x001FFFFD80000000L, *(ogg_int64_t*)(&doubletmp) - (ogg_int64_t)0x433FFFFD80000000L )

extern int   stop_decoding;
extern char  approxBRCaption[];
extern char  nomBRCaption[];
extern char  kCaption[];
extern char  speedCaption[];

dither_t   Dither;
double     doubletmp;

/*
 * Declare Functions
 */
static void usage(void);
double get_scale(oe_dec_opt *opt);

/*
 * Dither output
 */
ogg_int64_t dither_output(short dithering, long i, double Sum, long k, int format)
{
	short    shapingtype;
	double   Sum2;
	ogg_int64_t  val;

	if (dithering) {
		shapingtype = dithering - 1;
		if (!shapingtype) {
			double  tmp = Random_Equi ( Dither.Dither );
			Sum2 = tmp - Dither.LastRandomNumber [k];
			Dither.LastRandomNumber [k] = tmp;
			Sum2 = Sum += Sum2;
			val = ROUND64 (Sum2)  &  Dither.Mask;
		}
		else {
			Sum2  = Random_Triangular ( Dither.Dither ) - scalar16 ( Dither.DitherHistory[k], Dither.FilterCoeff + i );
			Sum  += Dither.DitherHistory [k] [(-1-i)&15] = Sum2;
			Sum2  = Sum + scalar16 ( Dither.ErrorHistory [k], Dither.FilterCoeff + i );
			val = ROUND64 (Sum2)  &  Dither.Mask;
			Dither.ErrorHistory [k] [(-1-i)&15] = (float)(Sum - val);
		}
	}
	else
		val = ((ogg_int64_t)(ROUND64 (Sum)));

	if (format == WAV_FMT_8BIT)
		val = val >> 24;
	else if (format == WAV_FMT_16BIT)
		val = val >> 16;
	else if (format == WAV_FMT_24BIT)
		val = val >> 8;

	return (val);
}

/* This downmixes multi-channel vorbis to stereo
 */
void downmix_2_stereo(int channels, float **pcm, long samples)
{
	/* Channels are in vorbis order NOT WAVEFORMATEXTENSIBLE */

	float **data = malloc(sizeof(float *) * 2);
	long n, ch = 0;

	for (ch = 0; ch < 2; ch++)
		data[ch] = malloc(sizeof(float) * samples);

	if (channels == 6) {
		for (n = 0; n < samples; n++) {
			data[0][n] = pcm[0][n] + pcm[1][n] * SQRTHALF + pcm[3][n] * SQRTHALF + pcm[5][n];
			data[1][n] = pcm[2][n] + pcm[1][n] * SQRTHALF + pcm[4][n] * SQRTHALF + pcm[5][n];
		}
	}
	else if (channels == 5) {
		for (n = 0; n < samples; n++) {
			data[0][n] = pcm[0][n] + pcm[1][n] * SQRTHALF + pcm[3][n] * SQRTHALF;
			data[1][n] = pcm[2][n] + pcm[1][n] * SQRTHALF + pcm[4][n] * SQRTHALF;
		}
	}
	else if (channels == 4) {
		for (n = 0; n < samples; n++) {
			data[0][n] = pcm[0][n] + pcm[2][n] * SQRTHALF;
			data[1][n] = pcm[1][n] + pcm[3][n] * SQRTHALF;
		}
	}
	else if (channels == 3) {
		for (n = 0; n < samples; n++) {
			data[0][n] = pcm[0][n] + pcm[1][n] * SQRTHALF;
			data[1][n] = pcm[2][n] + pcm[1][n] * SQRTHALF;
		}
	}

	for (ch = 0; ch < 2; ch++)
		for (n = 0; n < samples; n++)
			pcm[ch][n] = data[ch][n];

	for (ch = 0; ch < 2; ch++)
		free (data[ch]);
	free (data);

}

/* This is more or less the same as strncasecmp - but that doesn't exist
 * everywhere, and this is a fairly trivial function, so we include it 
 */
int tag_compare(const char *s1, const char *s2, int n)
{
	int c=0;
	while (c < n) {
		if (toupper(s1[c]) != toupper(s2[c]))
			return !0;
		c++;
	}
	return 0;
}

/* Reads VorbisGain tags from the input file. If audiophile is true, 
 * checks for audiophile tag, otherwise it reads the radio tag. If
 * the tags are found, it returns the scale value calculated from the
 * tag, otherwise, it returns a scale value of 1.0.
 */
double get_scale(oe_dec_opt *opt)
{
	vcedit_state   *state = NULL;
	vorbis_comment *vc;
	FILE           *file = NULL;
	int            i;
	unsigned int   j,
	               k;
	char           tag_buf[15];
	double         gain_A = 0.f,
	               gain_R = 0.f,
	               peak_A = 0.f,
	               peak_R = 0.f,
	               scale_factor = 1.0,
	               factor_clip;

	state = vcedit_new_state();

	if (state == NULL) {
		error_handler("Out of memory\n");
		goto exit;
	}

	file = fopen(opt->input_ogg, "rb");

	if (file == NULL) {
		error_handler("Couldn't open file\n");
		goto exit;
	}

	if (vcedit_open(state, file) < 0) {
		error_handler("Couldn't open file as vorbis\n");
		goto exit;
	}

	vc = vcedit_comments(state);

	for (i = 0; i < vc->comments; i++) { 
		/* Check for ReplayGain tags and get the peak and gain values. */
		if (!tag_compare(vc->user_comments[i], "RG_PEAK=", 8)) {
			for(j = 8, k = 0; j <= strlen(vc->user_comments[i]); j++, k++)
				tag_buf[k] = vc->user_comments[i][j];
			peak_R = atof(tag_buf);
			continue;
		}
		if (!tag_compare(vc->user_comments[i], "REPLAYGAIN_TRACK_PEAK=", 22)) {
			for(j = 22, k = 0; j <= strlen(vc->user_comments[i]); j++, k++)
				tag_buf[k] = vc->user_comments[i][j];
			peak_R = atof(tag_buf);
			continue;
		}
		if (!tag_compare(vc->user_comments[i], "RG_RADIO=", 9)) {
			for(j = 9, k = 0; j <= strlen(vc->user_comments[i]); j++, k++)
				tag_buf[k] = vc->user_comments[i][j];
			gain_R = atof(tag_buf);
			continue;
		}
		if (!tag_compare(vc->user_comments[i], "REPLAYGAIN_TRACK_GAIN=", 22)) {
			for(j = 22, k = 0; j <= strlen(vc->user_comments[i]); j++, k++)
				tag_buf[k] = vc->user_comments[i][j];
			gain_R = atof(tag_buf);
			continue;
		}
		if (!tag_compare(vc->user_comments[i], "RG_AUDIOPHILE=", 14)) {
			for(j = 14, k = 0; j <= strlen(vc->user_comments[i]); j++, k++)
				tag_buf[k] = vc->user_comments[i][j];
			gain_A = atof(tag_buf);
			continue;
		}
		if (!tag_compare(vc->user_comments[i], "REPLAYGAIN_ALBUM_GAIN=", 22)) {
			for(j = 22, k = 0; j <= strlen(vc->user_comments[i]); j++, k++)
				tag_buf[k] = vc->user_comments[i][j];
			gain_A = atof(tag_buf);
			continue;
		}
		if (!tag_compare(vc->user_comments[i], "REPLAYGAIN_ALBUM_PEAK=", 22)) {
			for(j = 22, k = 0; j <= strlen(vc->user_comments[i]); j++, k++)
				tag_buf[k] = vc->user_comments[i][j];
			peak_A = atof(tag_buf);
			continue;
		}
	}

	/* Do the scale calculations with clipping adjustment, if necessary. */
	if (opt->tagMode == 2) {           // audiophile
		if (peak_A > 0.f) {
			peak_A *= 32767.;
			scale_factor = pow(10., gain_A * 0.05);
			factor_clip  = (32767./( peak_A + 1));
			if (scale_factor < factor_clip)
				factor_clip = 1.;
			else
				factor_clip /= scale_factor;
			scale_factor *= factor_clip;
		}
		else if (gain_A > 0.f && peak_R > 0.f) {
			peak_R *= 32767.;
			scale_factor = pow(10., gain_A * 0.05);
			factor_clip  = (32767./( peak_R + 1));
			if (scale_factor < factor_clip)
				factor_clip = 1.;
			else
				factor_clip /= scale_factor;
			scale_factor *= factor_clip;
		}
		else
			if (gain_A > 0.f)
				scale_factor = pow(10., gain_A * 0.05);
	}
	if (opt->tagMode == 1) {           // radio
		if (peak_R > 0.f) {
			peak_R *= 32767.;
			scale_factor = pow(10., gain_R * 0.05);
			factor_clip  = (32767./( peak_R + 1));
			if (scale_factor < factor_clip)
				factor_clip = 1.;
			else
				factor_clip /= scale_factor;
			scale_factor *= factor_clip;
		}
		else
			if (gain_R > 0.f)
				scale_factor = pow(10., gain_R * 0.05);
	}

exit:
	if (state != NULL)
		vcedit_clear(state);
	if (file != NULL)
		fclose(file);

	return scale_factor;
}

/*
 * Main decoding function
 */
int oe_decode(oe_dec_opt *opt)
{
	OggVorbis_File    vf;
	int               current_section;
	vorbis_info       *vi;
	FILE              *ov_input;
	int               channels;
	audio_file        *aufile;
	void              *sample_buffer;
	long              bytesdone = 0,
	                  file_size;
	double            scale;
	int               l,
	                  eos = 0;
	int               dont_read_this_time = 0;
	unsigned int      section_1;
	unsigned int      section_2;
	double            time_diff;
	TIMER             *time_int;
	double            speed_ratio = 1.;
	float             nominalBitrate;

	if (opt->tagMode == 1 || opt->tagMode == 2)
		scale = get_scale(opt);
	else
		scale = 1.0;

	if ((ov_input = fopen(opt->input_ogg, "rb")) == NULL) {
		error_handler("\n ERROR: cannot open input file\n");
		return (0);
	}

	if (ov_open(ov_input, &vf, NULL, 0) < 0) {
		error_handler("\n Input does not appear to be an Ogg bitstream.\n");
		return (0);
	}

	file_size = ov_pcm_total(&vf,-1);
	bytesdone = 0;

	section_2 = 0;

	for(l = 0; l < ov_streams(&vf); l++) {
		eos = 0;

		vi=ov_info(&vf,-1);

		if (opt->downmix && vi->channels > 2 && vi->channels < 7)
			channels = 2;
		else
			channels = vi->channels;

		if (!opt->decode_mode) {
			/*
			 * Set up output to Windows playback routines
			 */
			if (Set_WIN_Params (INVALID_FILEDESC, vi->rate, SAMPLE_SIZE, channels) < 0) {
				error_handler("\nCan't access win32 play routines\n");
				return (0);
			}
		}
		else {
			/*
			 * Set up standard output to file
			 */
			if (ov_streams(&vf) > 1) {
				char tmp[8];
				char out_wave[MAX_PATH];
				int len;

				strcpy(out_wave, opt->out_wave);
				len = strlen(out_wave);
				while (out_wave[len] != '.') len--;

				if (out_wave[len] == '.')
					out_wave[len] = '\0';

				sprintf(tmp, "_%d.wav", l);
				strcat(out_wave, tmp);
				aufile = open_output_audio_file(out_wave, vi->rate, channels,
					opt->format, OUTPUT_WAV);
			}
			else
				aufile = open_output_audio_file(opt->out_wave, vi->rate, vi->channels,
					opt->format, OUTPUT_WAV);

			if (aufile == NULL) {
				error_handler("Not able to open output file: %s\n", opt->out_wave) ;
				ov_clear(&vf);
				return (0);
			}
		}

		strcpy(approxBRCaption, "Decoding-Ave.Bitrate");
		strcpy(kCaption, "");
		nominalBitrate = (float)ov_bitrate(&vf, l)/1000;
		sprintf(nomBRCaption, "%.1fkbps", nominalBitrate);

		/*
		 * get start time
		 */
		time_int = timer_start();

		/*
		 * Main decode loop - continue until end of file
		 */
		while (!eos) {
			float **pcm;
			long ret;

			if (stop_decoding)
				break;

			if (dont_read_this_time)
				dont_read_this_time = 0;
			else
				ret = ov_read_float(&vf, &pcm, READ_SIZE, &current_section);

			section_1 = current_section;
			if (section_1 != section_2) {
				section_2 = section_1;
				eos = 1;
				dont_read_this_time = 1;
			}
			else {

				bytesdone += ret;

				opt->progress_update(file_size, bytesdone);
				time_diff = timer_time(time_int);
				speed_ratio = opt->enc_speed(time_diff, vi->rate, bytesdone);
				sprintf(speedCaption, "%.3f x Realtime", speed_ratio);

				if (ret == 0) 
					/* EOF */
					eos = 1;
				else if (ret < 0) {
					/* error in the stream.  Not a problem, just reporting it in
					 * case we (the app) cares.  In this case, we don't. 
					 */
				}
				else {
					int   convsize=READ_SIZE;
					long  j,
					      i = 0,
					      k;
					long  bout=(ret<convsize?ret:convsize);

					if (opt->downmix && vi->channels > 2 && vi->channels < 7)
						downmix_2_stereo(vi->channels, pcm, bout);

					/* scale floats to 8, 16, 24 or 32 bit signed ints 
					 * (host order) (unless float output)
					 * and apply ReplayGain scaling, etc. 
					 */
					sample_buffer = malloc(sizeof(float)*channels*bout);
					for(k = 0; k < channels; k++) {
						for(j = 0; j < bout; j++, i++) {
							ogg_int64_t val;
							double Sum;
							pcm[k][j] *= (float)scale;
							if (opt->man_gain != 0)
								pcm[k][j] *= (float)pow(10., (double)opt->man_gain/20);
							if (opt->limiter) {	/* hard 6dB limiting */
								if (pcm[k][j] < -0.5)
									pcm[k][j] = (float)(tanh((pcm[k][j] + 0.5) / (1-0.5)) * (1-0.5) - 0.5);
								else if (pcm[k][j] > 0.5)
									pcm[k][j] = (float)(tanh((pcm[k][j] - 0.5) / (1-0.5)) * (1-0.5) + 0.5);
							}
							if (opt->format != WAV_FMT_FLOAT) {
								Sum = pcm[k][j]*2147483647.f;
								if (i > 31)
									i = 0;
								val = dither_output(opt->dither, i, Sum, k, opt->format);
								if (val > (ogg_int64_t)opt->wrap_prev_pos)
									val = (ogg_int64_t)opt->wrap_prev_pos;
								else if (val < (ogg_int64_t)opt->wrap_prev_neg)
									val = (ogg_int64_t)opt->wrap_prev_neg;
								pcm[k][j] = (float)val;
							}
							else {
								if (pcm[k][j] > opt->wrap_prev_pos)
									pcm[k][j] = opt->wrap_prev_pos;
								else if (pcm[k][j] < opt->wrap_prev_neg)
									pcm[k][j] = opt->wrap_prev_neg;
							}
						}
					}
					sample_buffer = output_to_PCM(pcm, sample_buffer, channels,
							bout, opt->format);
					if (!opt->decode_mode)         // playback
						WIN_Play_Samples (sample_buffer, sizeof(short) * channels * ret);
					else                          // write to file
						write_audio_file(aufile, sample_buffer, bout * channels);
					free(sample_buffer);
				}
			}
		}
		/*
		 * Finished with this file, so cleanup.
		 */
		if (opt->decode_mode == 0)
			WIN_Audio_close();
		else
			close_audio_file(aufile);
	}

	fclose(ov_input);
	ov_clear(&vf);
	if (opt->del_input)
		if (remove(opt->input_ogg) != 0)
			error_handler("Error deleting input file\n");

	return (0);
}

/*
 * end of decode.c
 */

