/* OggDropXPd - Filename: ENCODE.C
 *
 * Function: ogg encoder.
 *
 *   Copyright (c) 2002 - 2005 John Edwards <john.edwards33@ntlworld.com>
 *   Jack Moffit <jack@icecast.org> [as the original 'oggdrop' author]
 *
 *   Michael Smith <msmith@xiph.org> [as the author of the original oggenc]
 *
 *   Kenneth Arnold <kcarnold@yahoo.com> and 
 *   Christopher Montgomery <xiphmont@xiph.org>
 *   [as authors of code Mike borrowed from for oggenc]
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Library General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#include <vorbis/vorbisenc.h>
#include "audio.h"
#include "encode.h"
#include "misc.h"
#include "skeleton.h"

#define READSIZE 3072

extern int  stop_encoding;
extern char approxBRCaption[];
extern char nomBRCaption[];
extern char kCaption[];
extern char speedCaption[];

int oe_write_page(ogg_page *page, FILE *fp)
{
	int written;
	written = fwrite(page->header,1,page->header_len, fp);
	written += fwrite(page->body,1,page->body_len, fp);

	return written;
}

static void set_advanced_encoder_options(oe_enc_opt *opt, vorbis_info *vi)
{
	int manage = 0;
	struct ovectl_ratemanage2_arg ai;

	vorbis_encode_ctl(vi, OV_ECTL_RATEMANAGE2_GET, &ai);
	if (opt->use_damp == 1 && opt->bitrate_average_damping > 0.) {
		ai.bitrate_average_damping = opt->bitrate_average_damping;
		manage = 1;
	}
	if (opt->use_bias == 1) {
		if (opt->bitrate_limit_reservoir_bias >= 0.0 && opt->bitrate_limit_reservoir_bias <= 1.0) {
			ai.bitrate_limit_reservoir_bias = opt->bitrate_limit_reservoir_bias;
			manage = 1;
		}
	}
	if (opt->use_bits == 1) {
		if (opt->bitrate_limit_reservoir_bits >= 0) {
			ai.bitrate_limit_reservoir_bits = opt->bitrate_limit_reservoir_bits;
			manage = 1;
		}
	}
	if (opt->use_noise == 1) {
		if (opt->impulse_noisetune >= -15.0 && opt->impulse_noisetune <= 0.0)
			vorbis_encode_ctl(vi, OV_ECTL_IBLOCK_SET, &opt->impulse_noisetune);
	}
//	if (opt->use_trigger == 1) {
//		if (opt->impulse_trigger_profile >= 0.0 && opt->impulse_trigger_profile <= 6.0)
//			vorbis_encode_ctl(vi, OV_ECTL_GLOBAL_MAPPING_SET, &opt->impulse_trigger_profile);
//	}
	if (opt->lowpass != 0.0)
		vorbis_encode_ctl(vi, OV_ECTL_LOWPASS_SET, &opt->lowpass);

	if (manage) {
		ai.management_active = 1;
		if (vorbis_encode_ctl(vi, OV_ECTL_RATEMANAGE2_SET, &ai))
			error_handler("Failed to set advanced rate management parameters");
	}
}

void add_fishead_packet (ogg_stream_state *os) {

	fishead_packet fp;

	memset(&fp, 0, sizeof(fp));
	fp.ptime_n = 0;
	fp.ptime_d = 1000;
	fp.btime_n = 0;
	fp.btime_d = 1000;

	add_fishead_to_stream(os, &fp);
}

/*
 * Adds the fishead packets in the skeleton output stream along with the e_o_s packet
 */
void add_fisbone_packet (ogg_stream_state *os, oe_enc_opt *opt) {

	fisbone_packet fp;

	memset(&fp, 0, sizeof(fp));
	fp.serial_no = opt->serialno;
	fp.nr_header_packet = 3;
	fp.granule_rate_n = opt->rate;
	fp.granule_rate_d = 1;
	fp.start_granule = 0;
	fp.preroll = 2;
	fp.granule_shift = 0;

	add_message_header_field(&fp, "Content-Type", "audio/vorbis");

	add_fisbone_to_stream(os, &fp);
}

#ifdef HAVE_KATE
static void add_kate_fisbone_packet (ogg_stream_state *os, oe_enc_opt *opt, kate_info *ki) {

   fisbone_packet fp;

   memset(&fp, 0, sizeof(fp));
   fp.serial_no = opt->kate_serialno;
   fp.nr_header_packet = ki->num_headers;
   fp.granule_rate_n = ki->gps_numerator;
   fp.granule_rate_d = ki->gps_denominator;
   fp.start_granule = 0;
   fp.preroll = 0;
   fp.granule_shift = ki->granule_shift;

   add_message_header_field(&fp, "Content-Type", "application/x-kate");

   add_fisbone_to_stream(os, &fp);
}
#endif

#ifdef HAVE_KATE
static void add_kate_karaoke_style(kate_info *ki,unsigned char r,unsigned char g,unsigned char b,unsigned char a)
{
    kate_style *ks;
    int ret;

    if (!ki) return;

    ks=(kate_style*)malloc(sizeof(kate_style));
    kate_style_init(ks);
    ks->text_color.r = r;
    ks->text_color.g = g;
    ks->text_color.b = b;
    ks->text_color.a = a;
    ret=kate_info_add_style(ki,ks);
    if (ret<0) {
      fprintf(stderr, _("WARNING: failed to add Kate karaoke style\n"));
    }
}
#endif

/*
 * Main encoding function
 */

long oe_encode(oe_enc_opt *opt)
{
	ogg_stream_state os;
	ogg_stream_state so; /* stream for skeleton bitstream */
	ogg_stream_state ko; /* stream for kate bitstream */
	ogg_page 	 og;
	ogg_packet 	 op;

	vorbis_dsp_state vd;
	vorbis_block     vb;
	vorbis_info      vi;

#ifdef HAVE_KATE
    kate_info        ki;
    kate_comment     kc;
    kate_state       k;
    oe_lyrics        *lyrics=NULL;
    size_t           lyrics_index=0;
    double           vorbis_time = 0.0;
#endif

	long             samplesdone = 0,
	                 bytes_written = 0,
	                 packetsdone = 0;
	int              eos,
	                 ret = 0;
	double           time_diff;
	TIMER            *time_int;
	double           speed_ratio = 1.;
	double           nominalBitrate;

	float            **buffer;
	long             samples_read;

	if (opt->channels > 255) {
		error_handler("255 channels should be enough for anyone. (Sorry, vorbis doesn't support more)\n");
		return 1;
	}

	/* get start time
	 */
	time_int = timer_start();

	/* Have vorbisenc choose a mode for us
	 */
	vorbis_info_init(&vi);

	if (opt->oeMode == OE_MODE_QUALITY) {
		if (vorbis_encode_setup_vbr(&vi, opt->channels, opt->rate, opt->quality_coefficient)) {
			error_handler("Mode initialisation failed: invalid parameters for quality");
			vorbis_info_clear(&vi);
			return 1;
		}

		/* do we have optional hard bitrate restrictions? */
		if (opt->hard_q_limits && (opt->hard_max > 0 || opt->hard_min > 0)) {
			long bitrate;
			struct ovectl_ratemanage2_arg ai;

			vorbis_encode_ctl(&vi, OV_ECTL_RATEMANAGE2_GET, &ai);

			/* libvorbis 1.1 (and current svn) doesn't actually fill this in,
			 * which looks like a bug. It'll then reject it when we call the
			 * SET version below. So, fill it in with the values that libvorbis
			 * would have used to fill in this structure if we were using the
			 * bitrate-oriented setup functions. Unfortunately, some of those
			 * values are dependent on the bitrate, and libvorbis has no way to
			 * get a nominal bitrate from a quality value. Well, except by doing
			 * a full setup... So, we do that. 
			 * Also, note that this won't work correctly unless you have 
			 * version 1.1.1 or later of libvorbis.
			 */
			{
				vorbis_info vi2;
				vorbis_info_init(&vi2);
				vorbis_encode_setup_vbr(&vi2, opt->channels, opt->rate, opt->quality_coefficient);
				vorbis_encode_setup_init(&vi2);
				bitrate = vi2.bitrate_nominal;
				vorbis_info_clear(&vi2);
			}

			ai.bitrate_average_kbps = bitrate/1000;
			ai.bitrate_limit_reservoir_bits = bitrate * 2;
			ai.bitrate_average_damping = 1.5;
			ai.bitrate_limit_reservoir_bias = .1;
	
			/* And now the ones we actually wanted to set
			 */
			ai.bitrate_limit_min_kbps = opt->hard_min;
			ai.bitrate_limit_max_kbps = opt->hard_max;
			ai.management_active = 1;

			if (vorbis_encode_ctl(&vi, OV_ECTL_RATEMANAGE2_SET, &ai)) {
				error_handler("Failed to set bitrate hard min/max in quality mode");
				vorbis_info_clear(&vi);
				return 1;
			}
		}
	}
	else if (opt->oeMode == OE_MODE_VBR) {
		if (vorbis_encode_setup_managed(&vi, opt->channels, opt->rate, 
			opt->hard_max > 0 ? opt->hard_max * 1000 : -1,
			opt->vbr_bitrate * 1000,
			opt->hard_min > 0 ? opt->hard_min * 1000 : -1)) {
			error_handler("Mode initialisation failed: invalid parameters for VBR quality");
			vorbis_info_clear(&vi);
			return 1;
		}
	}
	else {
		if (vorbis_encode_setup_managed(&vi, opt->channels, opt->rate, 
			opt->max_bitrate > 0 ? opt->max_bitrate * 1000 : -1,
			opt->nom_bitrate * 1000,
			opt->min_bitrate > 0 ? opt->min_bitrate * 1000 : -1)) {

			error_handler("Mode initialisation failed: invalid parameters for bitrate");
			vorbis_info_clear(&vi);
			return 1;
		}
	}

	if (!opt->managed) {
		/* Turn off management entirely (if it was turned on).
		 */
		vorbis_encode_ctl(&vi, OV_ECTL_RATEMANAGE2_SET, NULL);
	}

	set_advanced_encoder_options(opt, &vi);
    
	vorbis_encode_setup_init(&vi);

	/* Now, set up the analysis engine, stream encoder, and other
	 * preparation before the encoding begins.
	 */

	vorbis_analysis_init(&vd,&vi);
	vorbis_block_init(&vd,&vb);

#ifdef HAVE_KATE
    if (opt->lyrics) {
        /* load lyrics */
        lyrics=load_lyrics(opt->lyrics);
        /* if it fails, don't do anything else for lyrics */
        if (!lyrics) {
            opt->lyrics = NULL;
        } else {
            /* init kate for encoding */
            kate_info_init(&ki);
            kate_info_set_category(&ki, "LRC");
            if (opt->lyrics_language)
              kate_info_set_language(&ki, opt->lyrics_language);
            else
              fprintf(stderr, _("WARNING: no language specified for %s\n"), opt->lyrics);
            kate_comment_init(&kc);
            kate_encode_init(&k,&ki);

            /* if we're in karaoke mode (we have syllable level timing info),
               add style info in case some graphical player is used */
            add_kate_karaoke_style(&ki, 255, 255, 255, 255);
            add_kate_karaoke_style(&ki, 255, 128, 128, 255);
        }
    }
#endif

	ogg_stream_init(&os, opt->serialno);
	if (opt->with_skeleton) 
		ogg_stream_init(&so, opt->skeleton_serialno);
	if (opt->lyrics)
        	ogg_stream_init(&ko, opt->kate_serialno);

	/* create the skeleton fishead packet and output it */ 
	if (opt->with_skeleton) { 
		add_fishead_packet(&so); 
		if ((ret = flush_ogg_stream_to_file(&so, opt->out))) {
			error_handler("Failed writing fishead packet to output stream\n");
			goto cleanup; 
		}
	}

	/* Now, build the three header packets and send through to the stream 
	 * output stage (but defer actual file output until the main encode loop)
	 */
	{
		ogg_packet header_main;
		ogg_packet header_comments;
		ogg_packet header_codebooks;
		int result;

		/* Build the packets
		 */
		vorbis_analysis_headerout(&vd,opt->comments,
				&header_main,&header_comments,&header_codebooks);

		/* And stream them out
		 */
		/* output the vorbis bos first, then the fisbone packets */
		ogg_stream_packetin(&os,&header_main);
		while((result = ogg_stream_flush(&os, &og)))
		{
			if(!result) break;
			ret = oe_write_page(&og, opt->out);
			if(ret != og.header_len + og.body_len)
			{
				error_handler("Failed writing header to output stream\n");
				ret = 1;
				goto cleanup; /* Bail and try to clean up stuff */
			}
		}

#ifdef HAVE_KATE
        if (opt->lyrics) {
            ogg_packet kate_op;
            ret = kate_ogg_encode_headers(&k, &kc, &kate_op);
            if (ret < 0) {
                opt->error(_("Failed encoding Kate header\n"));
                goto cleanup;
            }
            ogg_stream_packetin(&ko,&kate_op);
            while((result = ogg_stream_flush(&ko, &og)))
            {
                if(!result) break;
                ret = oe_write_page(&og, opt->out);
                if(ret != og.header_len + og.body_len)
                {
                    opt->error(_("Failed writing header to output stream\n"));
                    ret = 1;
                    goto cleanup; /* Bail and try to clean up stuff */
                }
            }
            ogg_packet_clear(&kate_op);
        }
#endif

		if (opt->with_skeleton) {
			add_fisbone_packet(&so, opt);
			if ((ret = flush_ogg_stream_to_file(&so, opt->out))) {
				error_handler("Failed writing fisbone header packet to output stream\n");
				goto cleanup; 
			}
#ifdef HAVE_KATE
            if (opt->lyrics) {
                add_kate_fisbone_packet(&so, opt, &ki);
                if ((ret = flush_ogg_stream_to_file(&so, opt->out))) {
                    opt->error(_("Failed writing fisbone header packet to output stream\n"));
                    goto cleanup;
                }
            }
#endif
		}

		/* write the next Vorbis headers */
		ogg_stream_packetin(&os,&header_comments);
		ogg_stream_packetin(&os,&header_codebooks);

		while ((result = ogg_stream_flush(&os, &og))) {
			if (!result) break;
			ret = oe_write_page(&og, opt->out);
			if (ret != og.header_len + og.body_len) {
				error_handler("Failed writing header to output stream\n");
				ret = 1;
				goto cleanup; /* Bail and try to clean up stuff */
			}
			else
				bytes_written += ret;
		}
	}

    /* build kate headers if requested */
#ifdef HAVE_KATE
    if (opt->lyrics) {
        while (kate_ogg_encode_headers(&k,&kc,&op)==0) {
          ogg_stream_packetin(&ko,&op);
          ogg_packet_clear(&op);
        }
        while((result = ogg_stream_flush(&ko, &og)))
        {
            if(!result) break;
            ret = oe_write_page(&og, opt->out);
            if(ret != og.header_len + og.body_len)
            {
                opt->error(_("Failed writing header to output stream\n"));
                ret = 1;
                goto cleanup; /* Bail and try to clean up stuff */
            }
        }
    }
#endif

	if (opt->with_skeleton) { 
		add_eos_packet_to_stream(&so); 
		if ((ret = flush_ogg_stream_to_file(&so, opt->out))) { 
			error_handler("Failed writing skeleton eos packet to output stream\n"); 
			goto cleanup;
		}
	}

	eos = 0;

	/* Main encode loop - continue until end of file
	 */
	while (!eos) {

		if (stop_encoding)
			break;

		buffer = vorbis_analysis_buffer(&vd, READSIZE);
		samples_read = opt->read_samples(opt->readdata, buffer, READSIZE);

		samplesdone += samples_read;

		/* Call progress update every 10 pages
		 */
		if (packetsdone % 10 == 0) {
			time_diff = timer_time(time_int);

			opt->progress_update(opt->total_samples_per_channel, samplesdone);
			speed_ratio = opt->enc_speed(time_diff, opt->rate, samplesdone);
		}

		/* Tell the library how many samples (per channel) we wrote 
		 * into the supplied buffer
		 */
		vorbis_analysis_wrote(&vd, samples_read);

		/* While we can get enough data from the library to analyse, one
		 * block at a time...
		 */
		while (vorbis_analysis_blockout(&vd,&vb)==1) {

			/* Do the main analysis, creating a packet
			 */
			vorbis_analysis(&vb,NULL);	
			vorbis_bitrate_addblock(&vb);

			while (vorbis_bitrate_flushpacket(&vd, &op)) {

				/* Add packet to bitstream
				 */
				ogg_stream_packetin(&os,&op);
				packetsdone++;

				/* If we've gone over a page boundary, we can do actual output,
				 * so do so (for however many pages are available)
				 */

				while (!eos) {
					int result = ogg_stream_pageout(&os,&og);
					if (!result)
						break;

					/* now that we have a new Vorbis page, we scan lyrics for any that is due */
#ifdef HAVE_KATE
					if (opt->lyrics && ogg_page_granulepos(&og)>=0) {
						vorbis_time = vorbis_granule_time(&vd, ogg_page_granulepos(&og));
						const oe_lyrics_item *item;
						while ((item = get_lyrics(lyrics, vorbis_time, &lyrics_index))) {
							ogg_packet kate_op;
							if (item->km) {
								ret = kate_encode_set_style_index(&k, 0);
								if (ret < 0) {
									opt->error(_("Failed encoding karaoke style - continuing anyway\n"));
								}
								ret = kate_encode_set_secondary_style_index(&k, 1);
								if (ret < 0) {
									opt->error(_("Failed encoding karaoke style - continuing anyway\n"));
								}
								ret = kate_encode_add_motion(&k, item->km, 0);
								if (ret < 0) {
									opt->error(_("Failed encoding karaoke motion - continuing anyway\n"));
								}
							}
							ret = kate_ogg_encode_text(&k, item->t0, item->t1, item->text, strlen(item->text)+1, &kate_op);
							if (ret < 0) {
								opt->error(_("Failed encoding lyrics - continuing anyway\n"));
							}
							else {
								ogg_stream_packetin(&ko, &kate_op);
								ogg_packet_clear(&kate_op);
								while (1) {
									ogg_page ogk;
									int result=ogg_stream_flush(&ko,&ogk);
									if (!result) break;
									ret = oe_write_page(&ogk, opt->out);
									if(ret != ogk.header_len + ogk.body_len)
									{
										opt->error(_("Failed writing data to output stream\n"));
										ret = 1;
										goto cleanup; /* Bail */
									}
									else
										bytes_written += ret;
								}
							}
						}
					}
#endif

					ret = oe_write_page(&og, opt->out);
					if (ret != og.header_len + og.body_len) {
						error_handler("Failed writing data to output stream\n");
						ret = 1;
						goto cleanup; /* Bail */
					}
					else
						bytes_written += ret; 

					if (ogg_page_eos(&og))
						eos = 1;
				}
			}
		}
		nominalBitrate = (double)(8.0*((double)bytes_written / ((double)samplesdone/(double)opt->rate)));
		strcpy(approxBRCaption, "Average Bitrate: ");
		sprintf(nomBRCaption, "%.1fkbps", (float)nominalBitrate/1000);
		if (opt->oeMode == OE_MODE_QUALITY)
			sprintf(kCaption, "-q %.2f | at ", (float)opt->quality_coefficient*10);
		else
			strcpy(kCaption, "");
		sprintf(speedCaption, "%.3f x Realtime", speed_ratio);
	}

	/* if encoding lyrics, signal EOS and cleanup the kate state */
#ifdef HAVE_KATE
	if (opt->lyrics) {
		ogg_packet kate_op;
		ret = kate_ogg_encode_finish(&k, vorbis_time, &kate_op);
		if (ret < 0) {
			opt->error(_("Failed encoding Kate EOS packet\n"));
		}
		else {
			ogg_stream_packetin(&ko,&kate_op);
			packetsdone++;
			ogg_packet_clear(&kate_op);

			eos = 0;
			while(!eos)
			{
				int result = ogg_stream_pageout(&ko,&og);
				if(!result) break;

				ret = oe_write_page(&og, opt->out);
				if(ret != og.header_len + og.body_len)
				{
					opt->error(_("Failed writing data to output stream\n"));
					ret = 1;
					goto cleanup; /* Bail */
				}
				else
					bytes_written += ret;

				if(ogg_page_eos(&og))
					eos = 1;
			}
		}
	}
#endif

	/*
	 * Cleanup time
	 */
cleanup:
	opt->ave_bitrate = nominalBitrate/1000;

#ifdef HAVE_KATE
    if (opt->lyrics) {
       ogg_stream_clear(&ko);
        kate_clear(&k);
        kate_info_clear(&ki);
        kate_comment_clear(&kc);
        free_lyrics(lyrics);
    }
#endif

    if (opt->with_skeleton)
        ogg_stream_clear(&so);

	ogg_stream_clear(&os);

	vorbis_block_clear(&vb);
	vorbis_dsp_clear(&vd);
	vorbis_info_clear(&vi);

	timer_clear(time_int);

	return bytes_written;
}

/*
 * end of encode.c
 */
