/* OggDropXPd - Filename: MISC.C
 *
 * Function: Miscellaneous functions for oggdropXPd.
 *
 *   Copyright (c) 2002 - 2005 John Edwards <john.edwards33@ntlworld.com>
 *   Jack Moffit <jack@icecast.org> [as the original 'oggdrop' author]
 *
 *   Michael Smith <msmith@xiph.org> [as the author of the original oggenc]
 *
 *   Kenneth Arnold <kcarnold@yahoo.com> and 
 *   Christopher Montgomery <xiphmont@xiph.org>
 *   [as authors of code Mike borrowed from for oggenc]
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Library General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <fcntl.h>
#include <io.h>
#include <stdlib.h>
#include <time.h>
#include <stdio.h>
#include <string.h>
#include <windows.h>
#include "misc.h"

static char *_filename;
void (*error_handler)(const char *fmt, ...) = error_dialog;


/*
 * For Bitrate Calculation for Quality settings
 */

double qc2approxBitrate(float qcValue, long rate, int channels)
//float qc2approxBitrate(int qcValue)
{

	vorbis_info vi;
	float br;

	vorbis_info_init(&vi);
	if (vorbis_encode_init_vbr(&vi, channels, rate, ((float)qcValue / 1000.0))) 
		return 128000.0; /* Mode setup failed: go with a default. */

	br = (float)vi.bitrate_nominal;
	vorbis_info_clear(&vi);

	return br;
/*
	float approxBitrate;

	if ( qcValue < 0) {
		approxBitrate = 64000. - 1000*((float)(((float)qcValue*-1./100.0)*19.));
	}
	else if ( qcValue < 410 ) {
		approxBitrate = 1000 * (float)(((float)qcValue/100.0)*16+64);
	}
	else if ( qcValue < 810 ) {
		approxBitrate = 1000 * (float)(((float)qcValue/100.0)*32);
	}
	else if ( qcValue < 910 ) {
		approxBitrate = 1000 *(float)((((float)qcValue/100.0)*32)+
						       ((((float)qcValue/100)-8)*32));
	}
	else {
		approxBitrate = 1000 * (float)((((float)qcValue/100.0)*32)+
	    		((((float)qcValue/100)-8)*32)+((((float)qcValue/100)-9)*116));
	}

	return approxBitrate;
*/
}

/*
 * would use real "strcasecmp" but it isn't portable
 */

int local_strcasecmp(const char* s1, const char* s2)
{
	unsigned char c1;
	unsigned char c2;
    
	do {
		c1 = tolower(*s1);
		c2 = tolower(*s2);
		if (!c1) {
			break;
		}
		++s1;
		++s2;
	} while (c1 == c2);
	return c1 - c2;
}

/*
 * Set the current input file name.
 */

void set_filename(char *filename)
{
	_filename = filename;
}

/*
 * Display an error dialog, possibly adding system error information.
 */

void error_dialog(const char *fmt, ...)
{
	va_list ap;
	char msgbuf[1024];
	char *bufp = msgbuf;

	/*
	 * A really rough sanity check to protect against blatant buffer overrun
	 */
	if (strlen(fmt) > 750)
		sprintf(msgbuf, "%s %s", "<buffer overflow> ", fmt);
	else {
		if (_filename != NULL && strlen(_filename) < 255) {
			sprintf(msgbuf, "%s: ", _filename);
			bufp += strlen(msgbuf);
		}

		va_start(ap, fmt);
		
		vsprintf(bufp, fmt, ap);

		va_end(ap);

		if (errno != 0) {
			bufp = msgbuf + strlen(msgbuf);
			sprintf(bufp, " error is %s (%d)", strerror(errno), errno);
			errno = 0;
		}
	}

	MessageBox(NULL, msgbuf, NULL, MB_OK | MB_ICONEXCLAMATION);
}

void log_error(const char *fmt, ...)
{
	va_list ap;
	FILE *fp;
	char msgbuf[1024];
	char *bufp = msgbuf;

	/*
	 * A really rough sanity check to protect against blatant buffer overrun
	 */
	if (strlen(fmt) > 750)
		sprintf(msgbuf, "%s %s", "<buffer overflow> ", fmt);
	else {
		va_start(ap, fmt);

		vsprintf(bufp, fmt, ap);

		va_end(ap);

		if (errno != 0) {
			bufp = msgbuf + strlen(msgbuf);
			sprintf(bufp, " error is: %s (%d)", strerror(errno), errno);
			errno = 0;
		}
	}

	va_start(ap, fmt);

	if ((fp = fopen("oggdropXPd.log", "a")) == (FILE *)NULL)
		return;

	fprintf(fp, "%s\n", msgbuf);
	fflush(fp);
	fclose(fp);

	va_end(ap);
}

void set_use_dialogs(int use_dialogs)
{
	if (!use_dialogs)
		error_handler = error_dialog;
	else
		error_handler = log_error;
}

void *timer_start(void)
{
	time_t *start = malloc(sizeof(time_t));
	time(start);
	return (void *)start;
}

double timer_time(void *time_int)
{
	time_t start = *((time_t *)time_int);
	time_t now = time(NULL);

	if (now-start)
		return (double)(now-start);
	else
		return 1;
}

void timer_clear(void *time_int)
{
	free((time_t *)time_int);
}

/*
 * end of misc.c
 */

