/* OggDropXPd - Filename: WAVE_OUT.C
 *
 * Function: Support for playing wave files - USE_WIN_AUDIO - ONLY.
 *
 *   Copyright (c) 2002 - 2005 John Edwards <john.edwards33@ntlworld.com>
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Library General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <windows.h>
#include <string.h>
#include <errno.h>
#include "audio.h"
#include "wave_out.h"

#define MAX_WAVEBLOCKS    32

#ifndef _WAVEFORMATEXTENSIBLE_
typedef struct {
    WAVEFORMATEX    outFormat;
    union {
        WORD wValidBitsPerSample;       /* bits of precision  */
        WORD wSamplesPerBlock;          /* valid if wBitsPerSample==0 */
        WORD wReserved;                 /* If neither applies, set to zero. */
    } Samples;
    DWORD           dwChannelMask;      /* which channels are */
                                        /* present in stream  */
    GUID            SubFormat;
} WAVEFORMATEXTENSIBLE, *PWAVEFORMATEXTENSIBLE;
#endif

static const GUID __KSDATAFORMAT_SUBTYPE_IEEE_FLOAT = {3, 0x0000, 0x0010, {0x80, 0x00, 0x00, 0xaa, 0x00, 0x38, 0x9b, 0x71}};
static const GUID __KSDATAFORMAT_SUBTYPE_PCM = {1, 0x0000, 0x0010, {0x80, 0x00, 0x00, 0xaa, 0x00, 0x38, 0x9b, 0x71}};


static CRITICAL_SECTION  cs;
static HWAVEOUT          dev                    = NULL;
static int               ScheduledBlocks        = 0;
static int               PlayedWaveHeadersCount = 0;          // free index
static WAVEHDR*          PlayedWaveHeaders [MAX_WAVEBLOCKS];


static int
Box ( const char* msg )
{
	MessageBox ( NULL, msg, NULL, MB_OK | MB_ICONEXCLAMATION );
	return -1;
}


/*
 *  This function registers already played WAVE chunks. Freeing is done by free_memory(),
 */

static void CALLBACK
wave_callback ( HWAVE hWave, UINT uMsg, DWORD dwInstance, DWORD dwParam1, DWORD dwParam2 )
{
	if ( uMsg == WOM_DONE ){
		EnterCriticalSection ( &cs );
		PlayedWaveHeaders [PlayedWaveHeadersCount++] = (WAVEHDR*) dwParam1;
		LeaveCriticalSection ( &cs );
	}
}


static void
free_memory ( void )
{
	WAVEHDR*  wh;
	HGLOBAL   hg;

	EnterCriticalSection ( &cs );
	wh = PlayedWaveHeaders [--PlayedWaveHeadersCount];
	ScheduledBlocks--;                        // decrease the number of USED blocks
	LeaveCriticalSection ( &cs );

	waveOutUnprepareHeader ( dev, wh, sizeof (WAVEHDR) );

	hg = GlobalHandle ( wh -> lpData );       // Deallocate the buffer memory
	GlobalUnlock (hg);
	GlobalFree   (hg);

	hg = GlobalHandle ( wh );                 // Deallocate the header memory
	GlobalUnlock (hg);
	GlobalFree   (hg);
}


Int
Set_WIN_Params ( FILE_T   dummyFile ,
                 Ldouble  SampleFreq,
                 Uint     BitsPerSample,
                 Uint     Channels )
{
	WAVEFORMATEX          outFormat;
	WAVEFORMATEXTENSIBLE  waveformat;
	UINT                  deviceID = WAVE_MAPPER;

	memset(&outFormat, 0, sizeof(WAVEFORMATEX));

	if ( waveOutGetNumDevs () == 0 )
		return Box ( "No audio device present." );

	if (Channels > 2) {
		waveformat.outFormat.wFormatTag = -2;
		waveformat.outFormat.cbSize = sizeof(WAVEFORMATEXTENSIBLE) - sizeof(WAVEFORMATEX);
		if (Channels == 3)
			waveformat.dwChannelMask = 7;
		else if (Channels == 4)
			waveformat.dwChannelMask = 51;
		else if (Channels == 5)
			waveformat.dwChannelMask = 55;
		else if (Channels == 6)
			waveformat.dwChannelMask = 63;
	}
	else {
		waveformat.outFormat.wFormatTag = WAVE_FORMAT_PCM;
	        waveformat.outFormat.cbSize = 0;
		waveformat.dwChannelMask = 0;
	}

	waveformat.outFormat.wBitsPerSample    = BitsPerSample;
	waveformat.outFormat.nChannels         = Channels;
	waveformat.outFormat.nSamplesPerSec    = (unsigned long)(SampleFreq + 0.5);
	waveformat.outFormat.nBlockAlign       = (waveformat.outFormat.wBitsPerSample + 7) / 8 * waveformat.outFormat.nChannels;
	waveformat.outFormat.nAvgBytesPerSec   = waveformat.outFormat.nSamplesPerSec * waveformat.outFormat.nBlockAlign;
	waveformat.Samples.wValidBitsPerSample = waveformat.outFormat.wBitsPerSample;
        waveformat.SubFormat                   = __KSDATAFORMAT_SUBTYPE_PCM;

	switch ( waveOutOpen ( &dev, deviceID, (WAVEFORMATEX *)&waveformat, (DWORD)wave_callback, 0, CALLBACK_FUNCTION ) ){
		case MMSYSERR_ALLOCATED:
			return Box ( "Device is already open." );
		case MMSYSERR_BADDEVICEID:
			return Box ( "The specified device is out of range." );
		case MMSYSERR_NODRIVER:
			return Box ( "There is no audio driver in this system." );
		case MMSYSERR_NOMEM:
			return Box ( "Unable to allocate sound memory." );
		case WAVERR_BADFORMAT:
			return Box ( "This audio format is not supported." );
		case WAVERR_SYNC:
			return Box ( "The device is synchronous." );
		default:
			return Box ( "Unknown media error." );
		case MMSYSERR_NOERROR:
			break;
	}

	waveOutReset ( dev );
	InitializeCriticalSection ( &cs );
	SetPriorityClass ( GetCurrentProcess (), HIGH_PRIORITY_CLASS );
	return 0;
}


int
WIN_Play_Samples ( const void* data, size_t len )
{
	HGLOBAL    hg;
	HGLOBAL    hg2;
	LPWAVEHDR  wh;
	void*      allocptr;

	do{
		while ( PlayedWaveHeadersCount > 0 )                        // free used blocks ...
			free_memory ();

		if ( ScheduledBlocks < sizeof(PlayedWaveHeaders)/sizeof(*PlayedWaveHeaders) ) // wait for a free block ...
			break;
		Sleep (26);

	} while (1);

	if ( (hg2 = GlobalAlloc ( GMEM_MOVEABLE, len )) == NULL )   // allocate some memory for a copy of the buffer
		return Box ( "GlobalAlloc failed." );

	allocptr = GlobalLock (hg2);
	CopyMemory ( allocptr, data, len );                         // Here we can call any modification output functions we want....

	if ( (hg = GlobalAlloc (GMEM_MOVEABLE | GMEM_ZEROINIT, sizeof (WAVEHDR))) == NULL ) // now make a header and WRITE IT!
		return -1;

	wh                   = GlobalLock (hg);
	wh->dwBufferLength   = len;
	wh->lpData           = allocptr;

	if ( waveOutPrepareHeader ( dev, wh, sizeof (WAVEHDR)) != MMSYSERR_NOERROR ){
		GlobalUnlock (hg);
		GlobalFree   (hg);
		return -1;
	}

	if ( waveOutWrite ( dev, wh, sizeof (WAVEHDR)) != MMSYSERR_NOERROR ){
		GlobalUnlock (hg);
		GlobalFree   (hg);
		return -1;
	}

	EnterCriticalSection ( &cs );
	ScheduledBlocks++;
	LeaveCriticalSection ( &cs );

	return len;
}


int
WIN_Audio_close ( void )
{
	if ( dev != NULL ){
		while ( ScheduledBlocks > 0 ){
			Sleep (ScheduledBlocks);
			while ( PlayedWaveHeadersCount > 0 )                        // free used blocks ...
			free_memory ();
		}

		waveOutReset (dev);      // reset the device
		waveOutClose (dev);      // close the device
		dev = NULL;
	}

	DeleteCriticalSection ( &cs );
	ScheduledBlocks = 0;
	return 0;
}

/*
 * end of wave_out.c
 */

